#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <sysexits.h>
#include <getopt.h>
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

void usage(const char *pname) {
	fprintf(stderr, "Usage: %s [-h|--help] [-b|--boardrsrc boardfile] [-d|--drvrrsrc driverfile] [-l|--bytelanes bytelanes] [-o|--output outfile]\n", pname);
	fprintf(stderr, "-h | --help:                 This usage statement\n");
	fprintf(stderr, "-f | --funcrsrc:             The filename of a functional sResource file\n");
	fprintf(stderr, "-b | --boardrsrc:            The filename of the board resource file\n");
	fprintf(stderr, "-l | --bytelanes:            The one byte hexadecimal value representing bytelanes\n");
	fprintf(stderr, "-o | --output:               The filename to write the resource to\n");
	return;
}

int main(int argc, char *argv[]) {
	char c = 0;
	int loptind = 0;
	char *boardfile = NULL;
	char *funcfile = NULL;
	char *outfile = NULL;
	uint8_t bytelanes = 0;
	int ifd;
	int ofd;
	struct option o[] = {
		{"help", 0, 0, 'h'},
		{"boardrsrc", 1, 0, 'b'},
		{"funcrsrc", 1, 0, 'f'},
		{"bytelanes", 1, 0, 'l'},
		{"output", 1, 0, 'o'},
		{0,0,0,0}
	};

	while( (c = getopt_long(argc, argv, "b:f:hl:o:", o, &loptind)) != -1 ) {
		switch(c) {
		case 'b':
			if(!optarg) {
				usage(argv[0]);
				exit(EX_USAGE);
			}
			boardfile = optarg;
			break;
		case 'f':
			if(!optarg) {
				usage(argv[0]);
				exit(EX_USAGE);
			}
			funcfile = optarg;
			break;
		case 'l':
			if(!optarg) {
				usage(argv[0]);
				exit(EX_USAGE);
			}
			sscanf(optarg, "%hhx", &bytelanes);
			break;
		case 'o':
			if(!optarg) {
				usage(argv[0]);
				exit(EX_USAGE);
			}
			outfile = optarg;
			break;
		case 'h':
			usage(argv[0]);
			exit(EX_USAGE);
			break;
		}
	}

	if(!boardfile || !outfile) {
		usage(argv[0]);
		exit(EX_USAGE);
	}
	errno = 0;

	struct stat boardstat;
	char *boardrsrc = NULL;
	ifd = open(boardfile, O_RDONLY);
	if(ifd < 0) {
		fprintf(stderr, "Could not open board file %s: %d %s\n", boardfile, errno, strerror(errno));
		exit(EX_CONFIG);
	}
	if(fstat(ifd, &boardstat) < 0) {
		fprintf(stderr, "Could not fstat board file %s: %d %s\n", boardfile, errno, strerror(errno));
		exit(EX_CONFIG);
	}
	boardrsrc = calloc(1, boardstat.st_size);
	if(!boardrsrc) exit(1);
	if(read(ifd, boardrsrc, boardstat.st_size) != boardstat.st_size) {
		fprintf(stderr, "Could not read board file %s: %d %s\n", boardfile, errno, strerror(errno));
		exit(EX_CONFIG);
	}
	close(ifd);

	struct stat funcstat;
	char *funcrsrc = NULL;
	if(funcfile) {
		ifd = open(funcfile, O_RDONLY);
		if(ifd < 0) {
			fprintf(stderr, "Could not open driver file %s: %d %s\n", funcfile, errno, strerror(errno));
			exit(EX_CONFIG);
		}
		if(fstat(ifd, &funcstat) < 0) {
			fprintf(stderr, "Could not fstat driver file %s: %d %s\n", funcfile, errno, strerror(errno));
			exit(EX_CONFIG);
		}
		funcrsrc = calloc(1, funcstat.st_size);
		if(!funcrsrc) exit(1);
		if(read(ifd, funcrsrc, funcstat.st_size) != funcstat.st_size) {
			fprintf(stderr, "Could not read driver file %s: %d %s\n", funcfile, errno, strerror(errno));
			exit(EX_CONFIG);
		}
		close(ifd);
	}

	ofd = open(outfile, O_WRONLY|O_CREAT|O_TRUNC, 0600);
	if(ofd < 0) {
		fprintf(stderr, "Could not open output file %s: %d %s\n", outfile, errno, strerror(errno));
		exit(EX_CONFIG);
	}

	uint32_t tmpint;
	uint32_t boardoffset = 4 /* ourselves */ + 4 /* EOL */;
	if(funcfile) boardoffset += 4;

	tmpint = 0x01000000 | boardoffset;
	tmpint = htonl(tmpint);
	if(write(ofd, &tmpint, sizeof(tmpint)) != sizeof(tmpint)) {
		fprintf(stderr, "Error writing board directory entry: %d %s\n", errno, strerror(errno));
		exit(EX_SOFTWARE);
	}
	
	if(funcfile) {
		uint32_t funcoffset = boardoffset - 4 + boardstat.st_size;
		if(funcrsrc) {
			tmpint = 0x80000000 | funcoffset;
			tmpint = htonl(tmpint);
			if(write(ofd, &tmpint, sizeof(tmpint)) != sizeof(tmpint)) {
				fprintf(stderr, "Error writing driver directory entry: %d %s\n", errno, strerror(errno));
				exit(EX_SOFTWARE);
			}
		}
	}

	tmpint = 0xFF000000;
	tmpint = htonl(tmpint);
	if(write(ofd, &tmpint, sizeof(tmpint)) != sizeof(tmpint)) {
		fprintf(stderr, "Error writing EOL directory entry: %d %s\n", errno, strerror(errno));
		exit(EX_SOFTWARE);
	}

	if(write(ofd, boardrsrc, boardstat.st_size) != boardstat.st_size) {
		fprintf(stderr, "Error writing board rsrc entry: %d %s\n", errno, strerror(errno));
		exit(EX_SOFTWARE);
	}

	if(funcfile) {
		if(write(ofd, funcrsrc, funcstat.st_size) != funcstat.st_size) {
			fprintf(stderr, "Error writing functional rsrc entry: %d %s\n", errno, strerror(errno));
			exit(EX_SOFTWARE);
		}
	}

	uint32_t len = lseek(ofd, 0, SEEK_CUR);
	tmpint = (-len) & 0x00FFFFFF;
	tmpint = htonl(tmpint);
	if(write(ofd, &tmpint, sizeof(tmpint)) != sizeof(tmpint)) {
		fprintf(stderr, "Error writing directory offset: %d %s\n", errno, strerror(errno));
		exit(EX_SOFTWARE);
	}

	len += 1 + 1 + 4 + 1 + 1 + 4 + 4 + 4; /* add format block to length */
	tmpint = htonl(len);
	if(write(ofd, &tmpint, sizeof(tmpint)) != sizeof(tmpint)) {
		fprintf(stderr, "Error writing length: %d %s\n", errno, strerror(errno));
		exit(EX_SOFTWARE);
	}

	tmpint = 0;
	if(write(ofd, &tmpint, sizeof(tmpint)) != sizeof(tmpint)) {
		fprintf(stderr, "Error writing empty crc: %d %s\n", errno, strerror(errno));
		exit(EX_SOFTWARE);
	}

	uint8_t tmpchar = 1;
	/* write revision level */
	if(write(ofd, &tmpchar, sizeof(tmpchar)) != sizeof(tmpchar)) {
		fprintf(stderr, "Error writing revision level: %d %s\n", errno, strerror(errno));
		exit(EX_SOFTWARE);
	}

	tmpchar = 1;
	if(write(ofd, &tmpchar, sizeof(tmpchar)) != sizeof(tmpchar)) {
		fprintf(stderr, "Error writing format: %d %s\n", errno, strerror(errno));
		exit(EX_SOFTWARE);
	}

	tmpint = htonl(0x5A932BC7);
	if(write(ofd, &tmpint, sizeof(tmpint)) != sizeof(tmpint)) {
		fprintf(stderr, "Error writing test pattern: %d %s\n", errno, strerror(errno));
		exit(EX_SOFTWARE);
	}

	tmpchar = 0;
	if(write(ofd, &tmpchar, sizeof(tmpchar)) != sizeof(tmpchar)) {
		fprintf(stderr, "Error writing reserved field: %d %s\n", errno, strerror(errno));
		exit(EX_SOFTWARE);
	}

	tmpchar = bytelanes;
	if(write(ofd, &tmpchar, sizeof(tmpchar)) != sizeof(tmpchar)) {
		fprintf(stderr, "Error writing bytelanes: %d %s\n", errno, strerror(errno));
		exit(EX_SOFTWARE);
	}

	close(ofd);
	exit(EX_OK);
}
